﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using PAT.Common.Classes.Expressions.ExpressionClass;

namespace PAT.Lib
{
    class SimpleTPM : DeviceBase
    {
        public const int REG_RET = 0;
        public const int REG_CMD = 1;

        public enum CMD { ExtendPCR, CreateEncryptKey, CertifyKey, LoadKey, UnloadKey, Decrypt };

        public const int PCRN = 2;
        //public const int LOADKEYN = 1;

        public const int SRTM = 0;
        public const int DRTM = 1;

        public List<PCR> pcr = null;

        public AKey ek_pri = null;
        public SKey srk = null;

        public AKey loadedkey_pri = null;
        public AKey loadedkey_pub = null;
        public int sealedpcrindex = -1;
        public Hash sealedpcr = null;

        public GDBase _ret = Value.Error();

        //reference to Cryptography
        public Cryptography crypt = null;

        public SimpleTPM()
        {
            pcr = new List<PCR>();
            for (int i = 0; i < PCRN; i++)
            {
                pcr.Add(new PCR());
            }
        }

        public SimpleTPM(SimpleTPM obj)
        {
            if (obj.ek_pri == null)
            {
                pcr = new List<PCR>();
                for (int i = 0; i < PCRN; i++)
                {
                    pcr.Add(new PCR());
                }
            }
            else
            {
                pcr = new List<PCR>();
                for (int i = 0; i < PCRN; i++) pcr.Add((PCR)obj.pcr[i].clone());
                ek_pri = (AKey)obj.ek_pri.clone();
                srk = (SKey)obj.srk.clone();
                if (obj.loadedkey_pri != null)
                {
                    loadedkey_pri = (AKey)obj.loadedkey_pri.clone();
                    loadedkey_pub = (AKey)obj.loadedkey_pub.clone();
                    sealedpcrindex = obj.sealedpcrindex;
                    sealedpcr = (Hash)obj.sealedpcr.clone();
                }
                _ret = obj._ret.clone();
                crypt = obj.crypt;
                crypt.tpm = this;
            }
        }

        public SimpleTPM(Cryptography _crypt, AKey _ek_pub, AKey _ek_pri, SKey _srk)
        {
            pcr = new List<PCR>();
            for (int i = 0; i < PCRN; i++)
            {
                pcr.Add(new PCR());
            }
            configureTPM(_crypt, _ek_pub, _ek_pri, _srk);
        }

        public override ExpressionValue GetClone()
        {
            return clone();
        }

        public override DeviceBase clone()
        {
            return new SimpleTPM(this);
        }

        public override string ToString()
        {
            string ret = "";
            if (ek_pri == null)
            {
                ret = "TPM Disabled";
            }
            else
            {
                ret += "TPM EKpri=" + ek_pri.ToString() + ",SRK=" + srk.ToString() + "\n";
                ret += " Ret=" + _ret.ToString() + "\n";
                for (int i = 0; i < PCRN; i++)
                {
                    ret += " PCR[" + i + "]:" + pcr[i].ToString();
                    if (i != PCRN - 1) ret += "\n";
                }
                if (loadedkey_pri != null)
                {
                    ret += "\n LOADKEY: (pub=" + loadedkey_pub.id + ",pri=" + loadedkey_pri.id + ") pcr[" + sealedpcrindex + "]=" + sealedpcr.ToString();
                }
            }
            return ret;
        }

        public override string ExpressionID
        {
            get
            {
                string ret = "";
                if (ek_pri == null)
                {
                    ret = "TPM Disabled";
                }
                else
                {
                    ret += "TPM\n";
                    ret += " EKpri=" + ek_pri.ExpressionID + ",SRK=" + srk.ExpressionID;
                    ret += " Ret=" + _ret.ExpressionID + "\n";
                    for (int i = 0; i < PCRN; i++)
                    {
                        ret += " PCR[" + i + "]:" + pcr[i].ExpressionID;
                        if (i != PCRN - 1) ret += "\n";
                    }
                    if (loadedkey_pri != null)
                    {
                        ret += " LOADKEY: (pub=" + loadedkey_pub.id + ",pri=" + loadedkey_pri.id + ") pcr[" + sealedpcrindex + "]=" + sealedpcr.ExpressionID;
                    }
                }
                return ret;
            }
        }

        public void configureTPM(Cryptography _crypt, AKey _ek_pub, AKey _ek_pri, SKey _srk)
        {
            crypt = _crypt;
            crypt.tpm = this;
            ek_pri = (AKey)_ek_pri.clone();
            srk = (SKey)_srk.clone();
            crypt.addKeyPair(_ek_pub.id, _ek_pri.id);
            for (int i = 0; i < PCRN; i++) pcr[i].plaintext = Value.Error();
        }

        public void invokeSRTM(GDBase bootloader)
        {
            pcr[SRTM] = new PCR(bootloader);
        }

        public void invokeDRTM(GDBase mle)
        {
            pcr[DRTM] = new PCR(mle);
        }

        public override GDBase read(int offset)
        {
            if (offset == REG_RET)
            {
                return _ret.clone();
            }
            else if (offset == REG_CMD)
            {
                //ignore
                return Value.Error();
            }
            else
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("tpm: invalid device read port");
        }

        public override void write(int offset, GDBase gd)
        {
            _ret = Value.Error();
            if (offset == REG_CMD)
            {
                if (gd is Concat)
                {
                    Concat cmd = (Concat)gd;
                    if (cmd.size >= 1 && cmd.component(0) is Value)
                    {
                        CMD cmdid = (CMD)((Value)cmd.component(0)).v;
                        switch (cmdid)
                        {
                            case CMD.ExtendPCR:
                                {
                                    if (cmd.size == 3 && cmd.component(1) is Value)
                                    {
                                        int pcrid = ((Value)cmd.component(1)).v;
                                        extendPCR(pcrid, cmd.component(2));
                                        _ret = new Value(0);
                                    }
                                    break;
                                }
                            case CMD.CreateEncryptKey:
                                {
                                    //CiphertextS createEncryptKey(int pcrindex, AKey key_pub, AKey key_pri)
                                    if (cmd.size == 4 && cmd.component(1) is Value && cmd.component(2) is AKey && cmd.component(3) is AKey)
                                    {
                                        int pcrid = ((Value)cmd.component(1)).v;
                                        _ret = createEncryptKey(pcrid, (AKey)cmd.component(2), (AKey)cmd.component(3));
                                    }
                                    break;
                                }
                            case CMD.LoadKey:
                                {
                                    //bool loadKey(AKey key_pub, CiphertextS sealedkey)
                                    if (cmd.size == 3 && cmd.component(1) is AKey && cmd.component(2) is CiphertextS)
                                    {
                                        if (loadKey((AKey)cmd.component(1), (CiphertextS)cmd.component(2)))
                                        {
                                            _ret = new Value(0);
                                        }
                                    }
                                    break;
                                }
                            case CMD.UnloadKey:
                                {
                                    //void unloadKey()
                                    if (cmd.size == 1 && isKeyLoaded())
                                    {
                                        unloadKey();
                                        _ret = new Value(0);
                                    }
                                    break;
                                }
                            case CMD.CertifyKey:
                                {
                                    //Signature certifyKey(Nonce nonce)
                                    if (cmd.size == 2 && cmd.component(1) is Nonce && isKeyLoaded())
                                    {
                                        _ret = certifyKey((Nonce)cmd.component(1));
                                    }
                                    break;
                                }
                            case CMD.Decrypt:
                                {
                                    //GDBase decrypt(CiphertextA ciphertext)
                                    if (cmd.size == 2 && cmd.component(1) is CiphertextA && isKeyLoaded())
                                    {
                                        _ret = decrypt((CiphertextA)cmd.component(1));
                                    }
                                    break;
                                }
                            default:
                                {
                                    _ret = Value.Error();
                                    break;
                                }
                        }
                    }
                }
            }
            else
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("tpm: invalid device write port");
        }

        public override int size()
        {
            return 2;
        }

        public override void reset()
        {
            pcr[SRTM] = new PCR(new Value(0));
            pcr[DRTM] = new PCR(new Value(-1));
            unloadKey();
        }

        private void extendPCR(int index, GDBase newdata)
        {
            pcr[index].extend(newdata);
        }

        private CiphertextS createEncryptKey(int pcrindex, AKey key_pub, AKey key_pri)
        {
            if (crypt == null) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("tpm: null Cryptography obj");
            crypt.addKeyPair(key_pub.id, key_pri.id);
            CiphertextS sealedkey = new CiphertextS(new Concat(key_pri, new Value(pcrindex), pcr[pcrindex]), srk);
            return sealedkey;
        }

        private Signature certifyKey(Nonce nonce)
        {
            //assert loaded already
            Concat cert = new Concat(loadedkey_pub, new Value(sealedpcrindex), sealedpcr, nonce);
            Signature sign = new Signature(cert, ek_pri);
            return sign;
            //return GenericData.List_2(cert, sign);
        }

        private bool loadableKey(AKey key_pub, CiphertextS sealedkey)
        {
            //throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException(sealedkey.ToString());
            if (crypt == null) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("tpm: null Cryptography obj");
            if (loadedkey_pri != null) return false;
            if (!(sealedkey.plaintext is Concat)) return false;
            Concat c = (Concat)sealedkey.plaintext;
            if (c.size != 3) return false;
            if (!(c.component(0) is AKey && c.component(1) is Value && c.component(2) is Hash)) return false;
            if (!crypt.isDecryptable(sealedkey, srk)) return false;
            if (!crypt.isKeyPair(key_pub, (AKey)c.component(0))) return false;
            //TODO: check
            //if (pcr[sealedkey.plaintext.list[1].csp_value].ExpressionID == sealedkey.plaintext.list[2].ExpressionID)
            return true;
            //else
            //   return false;
        }

        private bool loadKey(AKey key_pub, CiphertextS sealedkey)
        {
            if (crypt == null) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("tpm: null Cryptography obj");
            if (loadableKey(key_pub, sealedkey))
            {
                Concat c = (Concat)sealedkey.plaintext;
                loadedkey_pub = (AKey)key_pub.clone();
                loadedkey_pri = (AKey)(c.component(0)).clone();
                sealedpcrindex = ((Value)c.component(1)).v;
                sealedpcr = (Hash)(c.component(2)).clone();
                return true;
            }
            else
            {
                return false;
                //throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("cannot load the key");
            }
        }

        private void unloadKey()
        {
            loadedkey_pri = null;
        }

        private GDBase decrypt(CiphertextA ciphertext)
        {
            if (crypt == null) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("tpm: null Cryptography obj");
            GDBase ret = Value.Error();
            if (isKeyLoaded() && pcr[sealedpcrindex].ExpressionID == sealedpcr.ExpressionID)
            {
                if (crypt.isDecryptable(ciphertext, loadedkey_pri))
                {
                    ret = ciphertext.plaintext.clone();
                }
            }
            return ret;
        }

        public bool isKeyLoaded() { return loadedkey_pri != null; }


        //HACK
        public PCR getPCR(int idx)
        {
            if (pcr != null)
                return pcr[idx];
            else
                return null;
        }

    }
}
