﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;

namespace PAT.Lib
{
    class CiphertextS : GDBase
    {
        SKey _skey = null;
        GDBase _plaintext = null;

        public SKey skey
        {
            get { return (SKey)_skey.clone(); }
            set { _skey = (SKey)value.clone(); }
        }

        public GDBase plaintext
        {
            get { return _plaintext.clone(); }
            set { _plaintext = value.clone(); }
        }

        public CiphertextS()
        {
        }

        public CiphertextS(CiphertextS obj)
        {
            if (obj._skey != null)
            {
                _skey = (SKey)obj._skey.clone();
            }
            if (obj._plaintext != null)
            {
                _plaintext = obj._plaintext.clone();
            }
        }

        public CiphertextS(GDBase pl, SKey k)
        {
            plaintext = pl;
            skey = k;
        }

        public override ExpressionValue GetClone()
        {
            return clone();
        }

        public override GDBase clone()
        {
            return new CiphertextS(this);
        }

        public override string ToString()
        {
            return "cipherS(" + skey.ToString() + "," + plaintext.ToString() + ")";
        }

        public override string ExpressionID
        {
            get
            {
                return "cipherS(" + skey.ExpressionID + "," + plaintext.ExpressionID + ")";
            }
        }

        public override string typeString()
        {
            return "cipherS(" + plaintext.typeString() + ")";
        }

        public override List<GDBase> generate(List<GDBase> k)
        {
            List<GDBase> ret = base.generate(k);
            List<GDBase> keylist = skey.generate(k);
            if (keylist.Count > 0)
            {
                List<GDBase> plaintextlist = plaintext.generate(k);
                if (plaintextlist.Count > 0)
                {
                    foreach (SKey key in keylist)
                    {
                        foreach (GDBase gd in plaintextlist)
                        {
                            ret.Add(new CiphertextS(gd, key));
                        }
                    }
                }
            }
            return ret;
        }

        public override bool duplicate(List<GDBase> k)
        {
            bool ret = false;
            ret |= base.duplicate(k);
            ret |= skey.duplicate(k) && plaintext.duplicate(k);
            return ret;
        }
    }
}
