﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;

namespace PAT.Lib
{
    class Concat : GDBase
    {
        //int _id = -1;
        List<GDBase> _list = null;

        public List<GDBase> list { get { return _list; } }

        public Concat()
        {
        }

        public Concat(Concat obj)
        {
            if (obj._list != null)
            {
                _list = new List<GDBase>();
                foreach (GDBase gd in obj._list)
                {
                    _list.Add(gd.clone());
                }
            }
        }

        public Concat(params GDBase[] objs)
        {
            _list = new List<GDBase>();
            foreach (GDBase gd in objs)
            {
                _list.Add(gd.clone());
            }
        }

        public Concat(GDBase obj, Concat c)
        {
            _list = new List<GDBase>();
            _list.Add(obj.clone());
            foreach (GDBase gd in c.list)
            {
                _list.Add(gd.clone());
            }
        }

        public int size { get { return _list.Count; } }

        public GDBase component(int i)
        {
            return _list[i];
        }

        public void join(GDBase gd)
        {
            if (_list == null)
            {
                _list = new List<GDBase>();
            }
            _list.Add(gd.clone());
        }

        public void joinWithoutClone(GDBase gd)
        {
            if (_list == null)
            {
                _list = new List<GDBase>();
            }
            _list.Add(gd);
        }

        public override ExpressionValue GetClone()
        {
            return clone();
        }

        public override GDBase clone()
        {
            return new Concat(this);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("[");
            for (int i = 0; i < _list.Count; i++)
            {
                if (i != 0) sb.Append("||");
                sb.Append(_list[i].ToString());
            }
            sb.Append("]");
            return sb.ToString();
        }

        public override string ExpressionID
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                sb.Append("concat(");
                for (int i = 0; i < _list.Count; i++)
                {
                    if (i != 0) sb.Append(",");
                    sb.Append(_list[i].ExpressionID);
                }
                sb.Append(")");
                return sb.ToString();
            }
        }

        public override string typeString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("concat(");
            for (int i = 0; i < _list.Count; i++)
            {
                if (i != 0) sb.Append(",");
                sb.Append(_list[i].typeString());
            }
            sb.Append(")");
            return sb.ToString();
        }

        public override List<GDBase> generate(List<GDBase> k)
        {
            List<GDBase> ret = base.generate(k);
            List<GDBase>[] sublist = new List<GDBase>[_list.Count];
            bool failed = false;
            for (int i = 0; i < _list.Count; i++)
            {
                sublist[i] = _list[i].generate(k);
                if (sublist[i].Count == 0) failed = true;
            }
            if (!failed)
            {
                int[] index = new int[_list.Count];
                for (int i = 0; i < index.Length; i++) index[i] = 0;

               
                do
                {
                    Concat content = new Concat();
                    for (int i = 0; i < index.Length; i++)
                    {
                        content.joinWithoutClone(sublist[i][index[i]]);
                    }
                    ret.Add(content);
                    index[0]++;
                    for (int i = 0; i < index.Length - 1; i++)
                    {
                        if (index[i] == sublist[i].Count)
                        {
                            index[i] = 0;
                            index[i + 1]++;
                        }
                    }
                    if (index[index.Length - 1] == sublist[index.Length - 1].Count)
                        break;
                    else
                    {
                        //for (int i = 0; i < index.Length; i++) content._list[i] = sublist[i][index[i]];
                        //content._list[index.Length - 1] = sublist[index.Length - 1][index[index.Length - 1]];
                    }
                } while (true);
            }
            //remove duplicates

            //HashSet<string> set=new HashSet<string>();
            Dictionary<string, bool> dic = new Dictionary<string, bool>();

            for (int i = ret.Count - 1; i >= 0; i--)
            {
                string s = ret[i].ExpressionID;
                if (dic.ContainsKey(s))
                {
                    ret.RemoveAt(i);
                }
                else
                {
                    dic.Add(s, true);
                }
            }
            return ret;
        }

        public override bool duplicate(List<GDBase> k)
        {
            bool ret = false;
            ret |= base.duplicate(k);
            bool dupflag = true;
            foreach (GDBase gdb in list)
            {
                if (!gdb.duplicate(k))
                {
                    dupflag = false;
                    break;
                }
            }
            ret |= dupflag;
            return ret;
        }
    }
}
