﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;

namespace PAT.Lib
{
    class PagingTable : GDBase
    {
        public const int READ = 1;
        public const int WRITE = 2;
        public const int EXECUTE = 4;

        const int READ_ENCODE = 1024;
        const int WRITE_ENCODE = 2048;
        const int EXECUTE_ENCODE = 4096;

        List<int> _map = null;
        List<int> _ac = null;

        


        public List<int> map { get { return _map; } }
        public List<int> ac { get { return _ac; } }

        public PagingTable()
        {

        }

        public PagingTable(PagingTable obj)
        {
            if (obj.map != null)
            {
                _map = new List<int>();
                for (int i = 0; i < obj.map.Count; i++)
                {
                    _map.Add(obj._map[i]);
                }
            }
            if (obj._ac != null)
            {
                _ac = new List<int>();
                for (int i = 0; i < obj.ac.Count; i++)
                {
                    _ac.Add(obj._ac[i]);
                }
            }
        }

        public PagingTable(List<int> map, List<int> ac)
        {
            //TODO: assert they have the same size
            _map = map;
            _ac = ac;
        }
        
        public PagingTable(params int[] mapac)
        {
            //if (mapac.Length != Computer.MAXPAGE) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("invalid paging table");
            _map = new List<int>();
            _ac = new List<int>();
            for (int i = 0; i < mapac.Length; i++)
            {
                int encode = mapac[i];
                int mappedadr = encode & (READ_ENCODE - 1);
                int ac = 0;
                if ((encode & READ_ENCODE) != 0) ac += READ;
                if ((encode & WRITE_ENCODE) != 0) ac += WRITE;
                if ((encode & EXECUTE_ENCODE) != 0) ac += EXECUTE;
                _map.Add(mappedadr);
                _ac.Add(ac);
            }
        }

        public static PagingTable identityAllAccess(int pagenumber)
        {

            List<int> map = new List<int>();
            for (int i = 0; i < pagenumber; i++)
            {
                map.Add(0);
            }
            List<int> ac = new List<int>();
            for (int i = 0; i < pagenumber; i++)
            {
                map.Add(READ | WRITE | EXECUTE);
            }
            PagingTable pt = new PagingTable(map, ac);
            return pt;
        }

        public override ExpressionValue GetClone()
        {
            return clone();
        }

        public override GDBase clone()
        {
            return new PagingTable(this);
        }

        public override string ToString()
        {
            return ExpressionID;
        }

        public override string ExpressionID
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                sb.Append("PT(");
                for (int i = 0; i < map.Count; i++)
                {
                    if (i != 0) sb.Append(",");
                    sb.Append(map[i]);
                    sb.Append("[");
                    if ((ac[i] & READ) != 0) sb.Append("R");
                    if ((ac[i] & WRITE) != 0) sb.Append("W");
                    if ((ac[i] & EXECUTE) != 0) sb.Append("X");
                    sb.Append("]");
                }
                sb.Append(")");
                return sb.ToString();
            }
        }

        public override string typeString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("PT");
            return sb.ToString();
        }

        public override List<GDBase> generate(List<GDBase> k)
        {
            List<GDBase> ret = base.generate(k);
            ret.Add(PagingTable.identityAllAccess(map.Count));
            return ret;
        }

        public override bool duplicate(List<GDBase> k)
        {
            bool ret = false;
            ret |= base.duplicate(k);
            return ret;
        }
    }
}
