﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;

namespace PAT.Lib
{
    class Signature : GDBase
    {
        AKey _akey = null;
        GDBase _plaintext = null;

        public AKey akey
        {
            get { return (AKey)_akey.clone(); }
            set { _akey = (AKey)value.clone(); }
        }

        public GDBase plaintext
        {
            get { return _plaintext.clone(); }
            set { _plaintext = value.clone(); }
        }

        public Signature()
        {
        }

        public Signature(Signature obj)
        {
            if (obj._akey != null)
            {
                _akey = (AKey)obj._akey.clone();
            }
            if (obj._plaintext != null)
            {
                _plaintext = obj._plaintext.clone();
            }
        }

        public Signature(GDBase pl, AKey k)
        {
            plaintext = pl;
            akey = k;
        }

        public override ExpressionValue GetClone()
        {
            return clone();
        }

        public override GDBase clone()
        {
            return new Signature(this);
        }

        public override string ToString()
        {
            return ExpressionID;
        }

        public override string ExpressionID
        {
            get
            {
                return "sign(" + akey.ExpressionID + "," + plaintext.ExpressionID + ")";
            }
        }

        public override string typeString()
        {
            return "sign(" + plaintext.typeString() + ")";
        }

        public override List<GDBase> generate(List<GDBase> k)
        {
            List<GDBase> ret = base.generate(k);
            List<GDBase> keylist = akey.generate(k);
            if (keylist.Count > 0)
            {
                List<GDBase> plaintextlist = plaintext.generate(k);
                if (plaintextlist.Count > 0)
                {
                    foreach (AKey key in keylist)
                    {
                        foreach (GDBase gd in plaintextlist)
                        {
                            ret.Add(new Signature(gd, key));
                        }
                    }
                }
            }
            return ret;
        }

        public override bool duplicate(List<GDBase> k)
        {
            bool ret = false;
            ret |= base.duplicate(k);
            ret |= akey.duplicate(k) && plaintext.duplicate(k);
            return ret;
        }
    }
}
