﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using PAT.Common.Classes.Expressions.ExpressionClass;
namespace PAT.Lib
{
    partial class Computer : ExpressionValue
    {

        List<Processor> procs = null;    //Processors

        DeviceManager dm = new DeviceManager();       //Device Manager

        //Disk firstdisk = null;   //aux pointer to the first Disk
        //SimpleTPM tpm = null;   //aux pointer to TPM (only one is allowed)

        public Computer()
        {
        }

        public Computer(Computer obj)
        {
            if (obj.procs != null)
            {
                procs = new List<Processor>(obj.procs.Count);
                for (int i = 0; i < obj.procs.Count; i++)
                    procs.Add(new Processor(obj.procs[i]));
            }
            dm = new DeviceManager(obj.dm);

        }

        public override ExpressionValue GetClone()
        {
            return new Computer(this);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine();
            sb.Append(dm.ToString());
            if (procs != null)
            {
                for (int i = 0; i < procs.Count; i++)
                {
                    sb.Append("P[" + i + "]:");
                    sb.Append(procs[i].ToString());
                    sb.AppendLine();
                }
            }
            return sb.ToString();
        }

        public override string ExpressionID
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                sb.AppendLine();
                sb.Append(dm.ExpressionID);
                if (procs != null)
                {
                    for (int i = 0; i < procs.Count; i++)
                    {
                        sb.AppendLine();
                        sb.Append("P[" + i + "]:");
                        sb.Append(procs[i].ExpressionID);
                    }
                }
                return sb.ToString();
            }
        }


        //Machine Configurations
        public void configureProcessors(int count, int pagesize)
        {
            if (count <= 0 || pagesize <= 0)
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("invalid parameters");
            procs = new List<Processor>();
            for (int i = 0; i < count; i++)
            {
                procs.Add(new Processor(pagesize));
            }
        }

        public void addDevice(DeviceBase device, int baseadr)
        {
            dm.addDevice(device, baseadr);
        }

        //Machine-level Operations
        public void resetMachine()
        {
            //Reset processors
            for (int i = 0; i < procs.Count; i++)
            {
                procs[i].reset();
            }
            //Reset Devices
            dm.reset();
            //Activate BSP
            procs[0].activateLegacy();
        }

        public int procn()
        {
            if (procs != null)
                return procs.Count;
            else
                return 0;
        }

        public PCR getPCR(int idx)
        {
            if (dm.getTPM() != null)
                return dm.getTPM().getPCR(idx);
            else
                return null;
        }

        public int diskLen()
        {
            if (dm.getDisk() != null)
                return dm.getDisk().len();
            else
                return 0;
        }

        public void eavasdrop(int pid, Knowledge k)
        {
            checkLiveness(pid);
            if (procs[pid].inGuest)
            {
                int ptp_pa = procs[pid].guestPTP;
                if (readPA(ptp_pa) is PagingTable)
                {
                    PagingTable pt = (PagingTable)readPA(ptp_pa);
                    int ptsize = pt.ac.Count;
                    for (int vpage = 0; vpage < ptsize; vpage++)
                    {
                        int ac = pt.ac[vpage];
                        if ((ac & PagingTable.READ) != 0)
                        {
                            int pabase = pt.map[vpage] * procs[pid].pagesize;
                            for (int moffset = 0; moffset < procs[pid].pagesize; moffset++)
                            {
                                int pa = pabase + moffset;
                                if (dm.getDevice(pa) is RAM)
                                {
                                    k.addKnowledge(readPA(pa));
                                }
                            }
                        }

                    }
                }
                else
                    throw new SysError();
            }
            else
            {
                dm.eavasdropRAM(k);
            }
        }

        public bool nonStrongIsolation()
        {
            bool ret = false;
            if (procs == null) return ret;
            for (int i = 1; i < procs.Count - 1; i++)
            {
                if (procs[i].isActive && procs[i].inGuest)
                {
                    for (int j = i + 1; j < procs.Count; j++)
                    {
                        if (procs[j].isActive /*&& procs[j].inGuest*/)
                        {
                            try
                            {
                                int ptpi = procs[i].guestPTP;
                                int ptpj = procs[j].guestPTP;
                                if (readPA(ptpi) is PagingTable && readPA(ptpj) is PagingTable)
                                {
                                    PagingTable pti = (PagingTable)readPA(ptpi);
                                    PagingTable ptj = (PagingTable)readPA(ptpj);
                                    for (int x = 0; x < pti.map.Count; x++)
                                    {
                                        for (int y = 0; y < ptj.map.Count; y++)
                                        {
                                            if (pti.map[x] == ptj.map[y])
                                            {
                                                ret = true;
                                                break;
                                            }
                                        }
                                        if (ret) break;
                                    }
                                }
                            }
                            catch
                            {
                            }
                            if (ret) break;
                        }
                        if (ret) break;
                    }
                }
            }
            return ret;
        }

        public bool nonWeakIsolation()
        {
            bool ret = false;
            if (procs == null) return ret;
            for (int i = 1; i < procs.Count - 1; i++)
            {
                if (procs[i].isActive && procs[i].inGuest)
                {
                    for (int j = i + 1; j < procs.Count; j++)
                    {
                        if (procs[j].isActive /*&& procs[j].inGuest*/)
                        {
                            try
                            {
                                int ptpi = procs[i].guestPTP;
                                int ptpj = procs[j].guestPTP;
                                if (readPA(ptpi) is PagingTable && readPA(ptpj) is PagingTable)
                                {
                                    PagingTable pti = (PagingTable)readPA(ptpi);
                                    PagingTable ptj = (PagingTable)readPA(ptpj);
                                    for (int x = 0; x < pti.map.Count; x++)
                                    {
                                        for (int y = 0; y < ptj.map.Count; y++)
                                        {
                                            if (pti.map[x] == ptj.map[y] && (pti.ac[x] & PagingTable.WRITE) != 0 )
                                            {
                                                ret = true;
                                                break;
                                            }
                                        }
                                        if (ret) break;
                                    }
                                }
                            }
                            catch
                            {
                            }
                            if (ret) break;
                        }
                        if (ret) break;
                    }
                }
            }
            return ret;
        }
    }


}
