﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using PAT.Common.Classes.Expressions.ExpressionClass;
namespace PAT.Lib
{
    partial class Computer : ExpressionValue
    {



        private void VMEXitOrHalt(int pid)
        {
            if (procs[pid].inGuest)
                procs[pid].VMExit(Processor.VMExitReason.Error);
            else
                procs[pid].halt();
        }

        private void checkPID(int pid)
        {
            if (pid < 0 || pid >= procs.Count) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("illegal pid");
        }

        private void checkLiveness(int pid)
        {
            checkPID(pid);
            if (!procs[pid].isActive) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("illegal state processor " + pid);
        }

        private void checkSync(int pid)
        {
            checkLiveness(pid);
            if (!procs[pid].isSynchronized) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("VM[" + pid + "] is not synchronized. Reason=" + syncReason(pid) + " param=" + procs[pid].syncParam);
        }

        private void checkVA(int vadr)
        {
            if (vadr < 0) throw new SysError();
        }

        private void checkPA(int padr)
        {
            if (padr < 0) throw new SysError();
        }

        private int VA2PA(int pid, int vadr, out int ac)
        {
            checkLiveness(pid);
            checkVA(vadr);
            int pa = -1;
            if (procs[pid].inGuest)
            {
                int ptp_pa = procs[pid].guestPTP;
                if (readPA(ptp_pa) is PagingTable)
                {
                    PagingTable pt = (PagingTable)readPA(ptp_pa);
                    int pagesize = procs[pid].pagesize;
                    int vpage = vadr / pagesize;
                    int moffset = vadr % pagesize;
                    if (vpage < pt.map.Count && vpage < pt.ac.Count && pt.map[vpage] != -1)
                    {
                        pa = pt.map[vpage] * pagesize + moffset;
                        ac = pt.ac[vpage];
                    }
                    else
                    {
                        pa = -1;
                        ac = 0;
                    }
                }
                else
                    throw new SysError();
            }
            else
            {
                pa = vadr;
                ac = PagingTable.READ | PagingTable.WRITE | PagingTable.EXECUTE;
            }
            return pa;
        }

        private GDBase fetchMemByVA(int eid, int vadr, out int ac)
        {
            GDBase ret;
            int va = vadr;
            int padr = VA2PA(eid, va, out ac);
            ret = readPA(padr);
            return ret;
        }

        private GDBase readReg(int pid, int idx)
        {
            GDBase ret;
            ret = procs[pid].readReg(idx);
            return ret;
        }

        private void writeReg(int pid, int idx, GDBase data)
        {
            procs[pid].writeReg(idx, data.clone());
        }

        private GDBase decodeSrc(int pid, GDBase org, int deep = 0)
        {
            GDBase ret = null;
            if (org is Mem || org is RIP)
            {
                int ac;
                int va;
                if (org is Mem)
                {
                    va = ((Mem)org).adr;
                }
                else
                {
                    if (procs[pid].PP == -1) throw new SysError();
                    va = procs[pid].PP + ((RIP)org).offset;
                }
                //ret = fetchMemByVA(eid, va, out ac);
                int padr = VA2PA(pid, va, out ac);
                if ((ac & PagingTable.READ) == 0)
                    throw new SysError();
                ret = readPA(padr);
            }
            else if (org is Reg)
            {
                ret = readReg(pid, ((Reg)org).idx);
            }
            else if (org is Concat)
            {
                Concat oldlist = (Concat)org;
                Concat newlist = new Concat();
                for (int i = 0; i < oldlist.size; i++)
                {
                    newlist.join(decodeSrc(pid, oldlist.list[i], deep + 1));
                }
                ret = newlist;
            }
            else
            {
                //immediate operand
                ret = org;
            }
            return ret;
        }

        private void writeDst(int pid, GDBase org, GDBase data)
        {
            if (org is Mem || org is RIP)
            {
                int va;
                if (org is Mem)
                {
                    va = ((Mem)org).adr;
                }
                else
                {
                    if (procs[pid].PP == -1) throw new SysError();
                    va = procs[pid].PP + ((RIP)org).offset;
                }
                int ac;
                int padr = VA2PA(pid, va, out ac);
                if ((ac & PagingTable.WRITE) != 0)
                {
                    writePA(padr, data);
                }
                else
                    throw new SysError();
            }
            else if (org is Reg)
            {
                Reg reg = (Reg)org;
                writeReg(pid, reg.idx, data);
            }
            else
                throw new SysError();
        }

        //Access to Devices
        private GDBase readPA(int padr)
        {
            return dm.readPA(padr);
        }

        private void writePA(int padr, GDBase data)
        {
            dm.writePA(padr, data);
        }


        //invoked by CSP#

        //active
        public bool isActive(int pid) { return procs[pid].isActive; }

        public bool inLegacy(int pid) { return procs[pid].inLegacy; }
        public bool inGuest(int pid) { return procs[pid].inGuest; }
        public bool inHost(int pid) { return procs[pid].inHost; }


        public bool isAllInactive()
        {
            bool ret = true;
            for (int i = 0; i < procs.Count; i++)
                if (procs[i].isActive)
                {
                    ret = false;
                    break;
                }
            return ret;
        }

        //sync reason
        public int syncReason(int pid)
        {
            checkPID(pid);
            return procs[pid].syncID;
        }

        //clear synchronization
        public void synchronized(int pid)
        {
            checkPID(pid);
            procs[pid].synchronized();
        }

        //syncRunCode
        public int syncProg(int pid)
        {
            checkLiveness(pid);
            int codeid = procs[pid].cachedProgram.id;
            return codeid;
        }

        //syncChannelInput, syncChannelOutput
        public int syncChannel(int pid)
        {
            checkLiveness(pid);
            Value chn = (Value)procs[pid].syncParam.component(0);
            return chn.v;
        }

        //syncChannelIntput
        public GDBase syncDst(int pid)
        {
            checkLiveness(pid);
            GDBase dst = procs[pid].syncParam.component(1);
            return dst;
        }

        //syncChannelOutput
        public GDBase syncChannelData(int pid)
        {
            checkLiveness(pid);
            GDBase data = procs[pid].syncParam.component(1);
            return data;
        }

        //syncVMExit
        public int syncVMEXITReason(int pid)
        {
            checkLiveness(pid);
            return (int)procs[pid].vmexitReason;
        }

        //syncVMExit
        public int syncHypercall(int pid)
        {
            checkLiveness(pid);
            return procs[pid].hypercallID;
        }



        //fetch next program
        public void fetchNextProgram(int pid)
        {
            checkLiveness(pid);
            try
            {
                int va;
                int ac;
                procs[pid].synchronized();
                //checkSync(pid);
                va = procs[pid].nextPP;
                GDBase raw = fetchMemByVA(pid, va, out ac);
                if (((ac & PagingTable.READ) == 0) || ((ac & PagingTable.EXECUTE) == 0))
                    VMEXitOrHalt(pid);
                if (raw is Prog)
                {
                    Prog code = (Prog)raw;
                    procs[pid].cachedProgram = (Prog)code.clone();
                    procs[pid].syncPending(SyncException.syncRunCode());
                    procs[pid].PP = procs[pid].nextPP;
                    procs[pid].nextPP++;
                }
                else
                    VMEXitOrHalt(pid);
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }

        //eval
        public GDBase eval(int pid, GDBase src)
        {
            try
            {
                checkLiveness(pid);
                return decodeSrc(pid, src);
            }
            catch (SyncException)
            {
                return Value.Error();
            }
            catch (SysError)
            {
                return Value.Error();
            }
        }

        //copy received data to dst
        public void postChannelInput(int pid, GDBase chndata)
        {
            try
            {
                checkLiveness(pid);
                GDBase dst = syncDst(pid);
                procs[pid].synchronized();
                writeDst(pid, dst, chndata);
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }





        //instructions
        public void MOVE(int pid, GDBase dst, GDBase src)
        {
            //TODO: if trap MOVE
            GDBase srcData=null;
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                srcData = decodeSrc(pid, src);
                writeDst(pid, dst, srcData);
            }
            catch (SyncException se)
            {
                if (se.state == SyncException.SyncReason.ChannelInput)
                {
                    int chn = ((Value)se.param.component(0)).v;
                    procs[pid].syncPending(SyncException.syncChannelInput(chn, dst));
                }
                else if (se.state == SyncException.SyncReason.ChannelOutput)
                {
                    int chn = ((Value)se.param.component(0)).v;
                    procs[pid].syncPending(SyncException.syncChannelOutput(chn, se.param.component(1)));
                }
                else
                    procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }

        public void JUMP(int pid, GDBase src)
        {
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                GDBase srcData = decodeSrc(pid, src);
                if (srcData is Value)
                {
                    Value adr = (Value)srcData;
                    procs[pid].nextPP = adr.v;
                }
                else
                    throw new SysError();
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }

        public void HALT(int pid)
        {
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                if (procs[pid].inHost || procs[pid].inLegacy)
                    procs[pid].halt();
                else
                    throw SyncException.syncVMExit_PriviledgedInstruction();
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
            //TODO
            //vm[eid].syncPending(SyncException.syncHalt());
        }

        public void RELS(int pid)
        {
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                if (procs[pid].inHost)
                    procs[pid].VMEntry();
                else if (procs[pid].inGuest)
                    throw SyncException.syncVMExit_PriviledgedInstruction();
                else
                    throw new SysError();
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }

        public void WAKE(int pid, GDBase hventry, GDBase guestPTP, GDBase guestPNP)
        {
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                if (procs[pid].inHost || procs[pid].inLegacy)
                {
                    int newid = -1;
                    for (int i = 1; i < procs.Count; i++)
                    {
                        if (i == pid) continue;
                        if (!procs[i].isActive)
                        {
                            newid = i;
                            break;
                        }
                    }
                    if (newid == -1)
                        throw new SysError();
                    else
                    {
                        GDBase guestHVEntrygd = decodeSrc(pid, hventry);
                        GDBase guestPTPgd = decodeSrc(pid, guestPTP);
                        GDBase guestPNPgd = decodeSrc(pid, guestPNP);

                        if (guestHVEntrygd is Value && guestPTPgd is Value && guestPNPgd is Value)
                        {
                            int guestHVEntryi = ((Value)guestHVEntrygd).v;
                            int guestPTPi = ((Value)guestPTPgd).v;
                            int guestPNPi = ((Value)guestPNPgd).v;
                            //TODO: sanity check?
                            procs[newid].activateGuest(guestHVEntryi, guestPTPi, guestPNPi);
                        }
                        else
                            throw new SysError();
                    }
                }
                else if (procs[pid].inGuest)
                    throw SyncException.syncVMExit_PriviledgedInstruction();
                else
                    throw new SysError();
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }

        public void HYPC(int pid, GDBase callid)
        {
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                if (procs[pid].inGuest)
                {
                    GDBase callidgd = decodeSrc(pid, callid);
                    if (callidgd is Value)
                    {
                        int callidi = ((Value)callidgd).v;
                        //TODO: sanity check?
                        procs[pid].VMExit(Processor.VMExitReason.Hypercall, callidi);
                    }
                    else
                        throw new SysError();
                }
                else if (procs[pid].inGuest)
                    throw SyncException.syncVMExit_PriviledgedInstruction();
                else
                    throw new SysError();
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }
        }

        public void LL(int pid, GDBase pastart, GDBase len)
        {
            try
            {
                checkLiveness(pid);
                procs[pid].synchronized();
                if (procs[pid].inLegacy && pid==0)    
                {
                    //no other active processors
                    for (int i=1;i<procn();i++)
                        if (procs[i].isActive){
                            throw new SysError();
                        }
                    GDBase pastartgd = decodeSrc(pid, pastart);
                    GDBase lengd = decodeSrc(pid, len);
                    if (pastartgd is Value && lengd is Value)
                    {
                        int pastarti = ((Value)pastartgd).v;
                        int leni = ((Value)lengd).v;
                        //sanity check?
                        checkPA(pastarti);
                        if (leni > 0 && dm.getTPM() != null)
                        {
                            //DRTM
                            Concat mle = new Concat();
                            for (int i = 0; i < leni; i++)
                            {
                                mle.join(readPA(i + pastarti));
                            }
                            dm.getTPM().invokeDRTM(mle);
                            //partial reset 
                            procs[pid].activateLegacy(pastarti);
                            //sync
                            throw SyncException.syncLL();
                        }
                        else
                            throw new SysError();
                    }
                    else
                        throw new SysError();
                }
                else if (procs[pid].inGuest)
                    throw SyncException.syncVMExit_PriviledgedInstruction();
                else
                    throw new SysError();
            }
            catch (SyncException se)
            {
                procs[pid].syncPending(se);
            }
            catch (SysError)
            {
                VMEXitOrHalt(pid);
            }

        }



        //

    }
}