﻿using System;
using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;
namespace PAT.Lib
{
    //cryptography services. should be singleton
    class Cryptography : ExpressionValue
    {
        private Dictionary<int, int> keypair = new Dictionary<int, int>();

        public SimpleTPM tpm = null;

        public Cryptography()
        {

        }

        public Cryptography(Cryptography obj)
        {
            keypair = new Dictionary<int, int>(obj.keypair);
            if (obj.tpm != null)
            {
                tpm = obj.tpm;
                tpm.crypt = this;
            }
        }

        public void addKeyPair(int k1, int k2)
        {
            keypair.Add(k1, k2);
            keypair.Add(k2, k1);
            //throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException(k1.ToString() + " " + k2.ToString()+" "+ExpressionID);
        }

        public bool isDecryptable(GDBase gdb, GDBase key)
        {
            if (gdb is CiphertextA && key is AKey)
            {
                CiphertextA cipher = (CiphertextA)gdb;
                AKey akey = (AKey)key;
                int value;
                if (keypair.TryGetValue(akey.id, out value))
                    if (value == cipher.akey.id)
                        return true;
                    else
                        return false;
                else
                    return false;
            }
            else if (gdb is CiphertextS && key is SKey)
            {
                CiphertextS cipher = (CiphertextS)gdb;
                SKey skey = (SKey)key;
                return cipher.skey.id == skey.id;
            }
            else
            {
                return false;
            }
        }

        public bool verifySignature(GDBase gdb, GDBase key)
        {
            if (gdb is Signature && key is AKey)
            {
                Signature sign = (Signature)gdb;
                AKey akey = (AKey)key;
                int value;
                if (keypair.TryGetValue(akey.id, out value))
                    if (value == sign.akey.id)
                        return true;
                    else
                        return false;
                else
                    return false;
            }
            else
            {
                return false;
            }
        }


        public bool isKeyPair(GDBase key1, GDBase key2)
        {
            if (key1 is AKey && key2 is AKey)
            {
                AKey k1 = (AKey)key1;
                AKey k2 = (AKey)key2;
                int value;
                if (keypair.TryGetValue(k1.id, out value))
                    if (value == k2.id)
                        return true;
                    else
                        return false;
                else
                    return false;
            }
            else
                return false;
        }
        /*
        public GenericData verifyKeyCert(GenericData keycert, GenericData ek_pub, GenericData pcrindex, GenericData pcrvalue, GenericData nonce)
        {
            //assert types
            GenericData ret = GenericData.Value(-1);
            if (verifySignature(keycert, ek_pub))
            {
                GenericData cert = keycert.plaintext;
                if (cert.list[1].ExpressionID == pcrindex.ExpressionID && cert.list[2].ExpressionID == pcrvalue.ExpressionID && cert.list[3].ExpressionID == nonce.ExpressionID)
                {
                    ret = new GenericData(cert.list[0]);
                }

            }
            return ret;
        }
        */
        public override string ToString()
        {
            return "[" + ExpressionID + "]";
        }
        public override ExpressionValue GetClone()
        {
            return new Cryptography(this);
        }

        public override string ExpressionID
        {
            get
            {
                string ret = "";
                foreach (KeyValuePair<int, int> pair in keypair)
                {
                    if (pair.Key < pair.Value)
                        ret += "(" + pair.Key.ToString() + "," + pair.Value.ToString() + ")";
                }
                return ret;
            }
        }
    }
}
