﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using PAT.Common.Classes.Expressions.ExpressionClass;
namespace PAT.Lib
{
    class DeviceManager : ExpressionValue
    {
        List<DeviceBase> dev = new List<DeviceBase>();        //device (including memory)
        List<int> devba = new List<int>();             //device base address
        
        SimpleTPM tpm = null;
        Disk disk = null;

        public DeviceManager()
        {
        }

        public DeviceManager(DeviceManager obj)
        {
            for (int i = 0; i < obj.dev.Count; i++)
            {
                dev.Add(obj.dev[i].clone());
                if (dev[i] is SimpleTPM)
                    tpm = (SimpleTPM)dev[i];
                if (dev[i] is Disk && disk==null)
                    disk = (Disk)dev[i];
            }
            for (int i = 0; i < obj.devba.Count; i++)
                devba.Add(obj.devba[i]);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < dev.Count; i++)
            {
                sb.Append("VMEM[" + devba[i] + " - " + (dev[i].size() + devba[i] - 1) + "]: ");
                sb.AppendLine(dev[i].ToString());
            }
            return sb.ToString();
        }

        public override ExpressionValue GetClone()
        {
            return new DeviceManager(this);
        }

        public override string ExpressionID
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                for (int i = 0; i < dev.Count; i++)
                {
                    sb.Append("VMEM[" + devba[i] + " - " + (dev[i].size() + devba[i] - 1) + "]: ");
                    sb.AppendLine(dev[i].ExpressionID);
                }
                return sb.ToString();
            }
        }

        public bool addDevice(DeviceBase newdb, int baseadr)
        {
            bool succ = true;
            for (int i = 0; i < dev.Count; i++)
            {
                if ((baseadr >= devba[i] && baseadr < devba[i] + dev[i].size()) || (baseadr + newdb.size() - 1 >= devba[i] && baseadr + newdb.size() - 1 < devba[i] + dev[i].size()))
                {
                    succ = false;
                    break;
                }
            }
            if (newdb is SimpleTPM)
            {
                if (tpm != null)
                    throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("multiple TPMs");
                else
                    tpm = (SimpleTPM)newdb;
            }
            else if (newdb is Disk && disk==null)
            {
                disk = (Disk)newdb;
            }
            if (succ)
            {
                dev.Add(newdb);
                devba.Add(baseadr);
                return true;
            }
            else
            {
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("device memory mapping overlapped");
            }
        }

        //Read
        public GDBase readPA(int padr)
        {

            for (int i = 0; i < dev.Count; i++)
            {
                if (padr >= devba[i] && padr < devba[i] + dev[i].size())
                {
                    return dev[i].read(padr - devba[i]);
                }
            }
            throw new SysError();
        }

        //Write
        public void writePA(int padr, GDBase data)
        {
            for (int i = 0; i < dev.Count; i++)
            {
                if (padr >= devba[i] && padr < devba[i] + dev[i].size())
                {
                    dev[i].write(padr - devba[i], data);
                    return;
                }
            }
            throw new SysError();
        }

        //Reset
        public void reset()
        {
            for (int i = 0; i < dev.Count; i++) dev[i].reset();
        }

        public DeviceBase getDevice(int padr)
        {

            for (int i = 0; i < dev.Count; i++)
            {
                if (padr >= devba[i] && padr < devba[i] + dev[i].size())
                {
                    return dev[i];
                }
            }
            return null;
        }

        public SimpleTPM getTPM()
        {
            return tpm;
        }

        public Disk getDisk()
        {
            return disk;
        }

        public void eavasdropRAM(Knowledge k)
        {
            for (int i = 0; i < dev.Count; i++)
            {
                if (dev[i] is RAM)
                {
                    RAM ram = (RAM)dev[i];
                    for (int j = 0; j < ram.size(); j++)
                    {
                        k.addKnowledge(ram.read(j));
                    }
                }
            }
        }
    }
}
