 using System;
using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;

//the namespace must be PAT.Lib, the class and method names can be arbitrary
namespace PAT.Lib
{
    /// <summary>
    /// </summary>

    public class GenericData : ExpressionValue, IEquatable<GenericData>
    {

        /// <summary>
        ///
        /// </summary>
        /// <returns></returns>
        public enum TYPE { value, reg, mem, codeseg, akey, skey, list, ciphertextbyak, ciphertextbysk, signature, hash, xor, boolean, nonce, otp, var };
        public TYPE type;

        //name
        //public int name = -1;

        //value
        public int csp_value;
        //codeseg
        public int csp_code;
        //key
        public int csp_key;

        //bool
        public bool csp_bool;

        //list, xor
        public List<GenericData> list = null;
        //ciphertext, signature, hash
        public GenericData plaintext = null;


        public static GenericData Value(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.value;
            gd.csp_value = i;
            return gd;
        }

        public static GenericData Mem(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.mem;
            gd.csp_value = i;
            return gd;
        }

        public static GenericData Reg(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.reg;
            gd.csp_value = i;
            return gd;
        }

        public static GenericData CodeSeg(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.codeseg;
            gd.csp_code = i;
            return gd;
        }

        public static GenericData AKey(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.akey;
            gd.csp_key = i;
            return gd;
        }

        public static GenericData SKey(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.skey;
            gd.csp_key = i;
            return gd;
        }

        public static GenericData Bool(bool i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.boolean;
            gd.csp_bool = i;
            return gd;
        }

        public static GenericData Nonce(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.nonce;
            gd.csp_value = i;
            return gd;
        }

        public static GenericData OTP(int i)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.otp;
            gd.csp_value = i;
            return gd;
        }

        /*
        public static GenericData Ciphertext(GenericData _plaintext, int _key)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.ciphertext;
            gd.plaintext = _plaintext;
            gd.csp_key = _key;
            return gd;
        }
        */
        public static GenericData CiphertextA(GenericData _plaintext, GenericData _key)
        {
            if (_key.type != TYPE.akey)
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("not an asym key");
            else
            {
                GenericData gd = new GenericData();
                gd.type = GenericData.TYPE.ciphertextbyak;
                gd.plaintext = _plaintext;
                gd.csp_key = _key.csp_key;
                return gd;
            }
        }

        public static GenericData CiphertextS(GenericData _plaintext, GenericData _key)
        {
            if (_key.type != TYPE.skey)
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("not a sym key");
            else
            {
                GenericData gd = new GenericData();
                gd.type = GenericData.TYPE.ciphertextbysk;
                gd.plaintext = _plaintext;
                gd.csp_key = _key.csp_key;
                return gd;
            }
        }
        /*
        public static GenericData List(List<GenericData> gds)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            foreach (GenericData data in gds) gd.list.Add(new GenericData(data));
            return gd;
        }
        */
        public static GenericData List(GenericData[] gds)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            foreach (GenericData data in gds) gd.list.Add(new GenericData(data));
            return gd;
        }

        public static GenericData List_1(GenericData gd1)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            return gd;
        }

        public static GenericData List_2(GenericData gd1, GenericData gd2)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            return gd;
        }

        public static GenericData List_3(GenericData gd1, GenericData gd2, GenericData gd3)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            gd.list.Add(new GenericData(gd3));
            return gd;
        }

        public static GenericData List_4(GenericData gd1, GenericData gd2, GenericData gd3, GenericData gd4)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            gd.list.Add(new GenericData(gd3));
            gd.list.Add(new GenericData(gd4));
            return gd;
        }

        public static GenericData List_5(GenericData gd1, GenericData gd2, GenericData gd3, GenericData gd4, GenericData gd5)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            gd.list.Add(new GenericData(gd3));
            gd.list.Add(new GenericData(gd4));
            gd.list.Add(new GenericData(gd5));
            return gd;
        }

        public static GenericData List_6(GenericData gd1, GenericData gd2, GenericData gd3, GenericData gd4, GenericData gd5, GenericData gd6)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            gd.list.Add(new GenericData(gd3));
            gd.list.Add(new GenericData(gd4));
            gd.list.Add(new GenericData(gd5));
            gd.list.Add(new GenericData(gd6));
            return gd;
        }

        public static GenericData List_7(GenericData gd1, GenericData gd2, GenericData gd3, GenericData gd4, GenericData gd5, GenericData gd6, GenericData gd7)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.list;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            gd.list.Add(new GenericData(gd3));
            gd.list.Add(new GenericData(gd4));
            gd.list.Add(new GenericData(gd5));
            gd.list.Add(new GenericData(gd6));
            gd.list.Add(new GenericData(gd7));
            return gd;
        }

        public static GenericData List_n(int n, GenericData gd)
        {
            GenericData ret = new GenericData();
            ret.type = GenericData.TYPE.list;
            ret.list = new List<GenericData>();
            for (int i = 0; i < n; i++) ret.list.Add(new GenericData(gd));
            return ret;
        }

        public static GenericData ListVar(int n, params GenericData[] gds)
        {
            GenericData ret = new GenericData();
            ret.type = GenericData.TYPE.list;
            ret.list = new List<GenericData>();
            foreach (GenericData gd in gds)
            {
                ret.list.Add(new GenericData(gd));
            }
            return ret;
        }

        public static GenericData Xor(GenericData[] gds)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.xor;
            gd.list = new List<GenericData>();
            foreach (GenericData data in gds) gd.list.Add(new GenericData(data));
            //TODO: 
            //reduce A xor B xor A
            //reduce if element is xor structure
            return gd;
        }

        public static GenericData Xor_2(GenericData gd1, GenericData gd2)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.xor;
            gd.list = new List<GenericData>();
            gd.list.Add(new GenericData(gd1));
            gd.list.Add(new GenericData(gd2));
            //foreach (GenericData data in gds) gd.list.Add(new GenericData(data));
            //TODO: 
            //reduce A xor B xor A
            //reduce if element is xor structure
            return gd;
        }

        public static GenericData Signature(GenericData _plaintext, GenericData _key)
        {
            if (_key.type != TYPE.akey)
                throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("not a key");
            else
            {
                GenericData gd = new GenericData();
                gd.type = GenericData.TYPE.signature;
                gd.plaintext = _plaintext;
                gd.csp_key = _key.csp_key;
                return gd;
            }
        }

        public static GenericData Hash(GenericData _plaintext)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.hash;
            gd.plaintext = _plaintext;
            return gd;
        }

        public static GenericData Var(int id)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.var;
            gd.csp_value = id;
            return gd;
        }
        /*
        public static GenericData Result(GenericData code, GenericData input)
        {
            GenericData gd = new GenericData();
            gd.type = GenericData.TYPE.ret;
            gd.csp_code = code.csp_code;
            gd.plaintext = new GenericData(input);
            return gd;
        }
        */
        public GenericData()
        {
            type = TYPE.value;
            csp_value = -1;
        }

        public bool isReg { get { return type == TYPE.mem; } }
        /*
        public void setName(int _name)
        {
            name = _name;
        }

        public void clearName()
        {
            name = -1;
        }
        */
        public GenericData component(int i)
        {
            GenericData ret = GenericData.Value(-1);
            if (type == TYPE.list && i < list.Count) ret = new GenericData(list[i]);
            return ret;
        }

        public GenericData clear()
        {
            return new GenericData();
        }

        public GenericData(GenericData gd)
        {
            type = gd.type;
            //name = gd.name;
            switch (gd.type)
            {
                case TYPE.value:
                    csp_value = gd.csp_value;
                    break;
                case TYPE.mem:
                    csp_value = gd.csp_value;
                    break;
                case TYPE.reg:
                    csp_value = gd.csp_value;
                    break;
                case TYPE.codeseg:
                    csp_code = gd.csp_code;
                    break;
                case TYPE.akey:
                    csp_key = gd.csp_key;
                    break;
                case TYPE.skey:
                    csp_key = gd.csp_key;
                    break;
                case TYPE.list:
                    list = new List<GenericData>();
                    foreach (GenericData gde in gd.list)
                    {
                        list.Add(new GenericData(gde));
                    }
                    break;
                case TYPE.ciphertextbyak:
                    plaintext = new GenericData(gd.plaintext);
                    csp_key = gd.csp_key;
                    break;
                case TYPE.ciphertextbysk:
                    plaintext = new GenericData(gd.plaintext);
                    csp_key = gd.csp_key;
                    break;
                case TYPE.signature:
                    plaintext = new GenericData(gd.plaintext);
                    csp_key = gd.csp_key;
                    break;
                case TYPE.hash:
                    plaintext = new GenericData(gd.plaintext);
                    break;
                case TYPE.xor:
                    list = new List<GenericData>();
                    foreach (GenericData gde in gd.list)
                    {
                        list.Add(new GenericData(gde));
                    }
                    break;
                case TYPE.boolean:
                    csp_bool = gd.csp_bool;
                    break;
                case TYPE.nonce:
                    csp_value = gd.csp_value;
                    break;
                case TYPE.otp:
                    csp_value = gd.csp_value;
                    break;
                case TYPE.var:
                    csp_value = gd.csp_value;
                    break;
                /*
            case TYPE.ret:
                plaintext = new GenericData(gd.plaintext);
                csp_code = gd.csp_code;
                break;*/
                default:
                    throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("Unknown type");
            }
        }

        public override string ToString()
        {
            return "[" + ExpressionID + "]";
        }


        /// <summary>
        /// Please implement this method to return a deep clone of the current object
        /// </summary>
        /// <returns></returns>
        public override ExpressionValue GetClone()
        {
            return new GenericData(this);
        }

        public override string ExpressionID
        {
            get
            {
                string ret = "";
                switch (type)
                {
                    case TYPE.value:
                        ret = csp_value.ToString();
                        break;
                    case TYPE.mem:
                        ret = "m(" + csp_value.ToString() + ")";
                        break;
                    case TYPE.reg:
                        ret = "r(" + csp_value.ToString() + ")";
                        break;
                    case TYPE.codeseg:
                        ret = "c(" + csp_code.ToString() + ")";
                        break;
                    case TYPE.akey:
                        ret = "ak(" + csp_key.ToString() + ")";
                        break;
                    case TYPE.skey:
                        ret = "sk(" + csp_key.ToString() + ")";
                        break;
                    case TYPE.list:
                        {//Ordered
                            ret = "{";
                            //       bool first = true;
                            foreach (GenericData gd in list)
                            {
                                //     if (!first) ret += ",";
                                ret += gd.ExpressionID;
                                //      first = false;
                                ret += "||";
                            }
                            ret = ret.Substring(0, ret.Length - 2);
                            ret += "}";
                            break;
                        }
                    case TYPE.ciphertextbyak:
                        ret = "ciphera(" + plaintext.ExpressionID + ",ak(" + csp_key + "))";
                        break;
                    case TYPE.ciphertextbysk:
                        ret = "ciphers(" + plaintext.ExpressionID + ",sk(" + csp_key + "))";
                        break;
                    case TYPE.signature:
                        ret = "sign(" + plaintext.ExpressionID + ",ak(" + csp_key + "))";
                        break;
                    case TYPE.hash:
                        ret = "hash(" + plaintext.ExpressionID + ")";
                        break;
                    case TYPE.xor:
                        {//Unordered
                            ret = "xor(";
                            List<string> sublist = new List<string>();
                            foreach (GenericData gd in list)
                                sublist.Add(gd.ExpressionID);
                            sublist.Sort();
                            bool first = true;
                            foreach (string s in sublist)
                            {
                                if (!first) ret += ",";
                                ret += s;
                                first = false;
                            }
                            ret += ")";
                            break;
                        }
                    case TYPE.boolean:
                        ret = "b(" + csp_bool.ToString() + ")";
                        break;
                    case TYPE.nonce:
                        ret = "n(" + csp_value + ")";
                        break;
                    case TYPE.otp:
                        ret = "o(" + csp_value + ")";
                        break;
                    case TYPE.var:
                        ret = "var(" + csp_value + ")";
                        break;
                    /*
                case TYPE.ret:
                    ret = "f(" + csp_code+","+plaintext.ExpressionID + ")";
                    break;*/
                    default:
                        throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("Unknown type");
                }
                return ret;
            }
        }

        public bool Equals(GenericData other)
        {
            if (ExpressionID == other.ExpressionID)
                return true;
            else
                return false;
        }

        public GenericData opXor(GenericData gd)
        {
            if (type == TYPE.xor)
            {
                GenericData ret = new GenericData(this);
                if (list.Contains(gd))
                {
                    ret.list.Remove(gd);
                    if (ret.list.Count == 0)
                    {
                        ret = GenericData.Value(0);
                    }
                    else if (ret.list.Count == 1)
                    {
                        ret = new GenericData(ret.list[0]);
                    }
                }
                else
                {
                    ret.list.Add(gd);
                }
                return ret;
            }
            else
            {
                return Xor_2(this, gd);
            }
        }

        public string TypeString
        {
            get
            {
                string ret = "";
                switch (type)
                {
                    case TYPE.value:
                        ret = "i";
                        break;
                    case TYPE.mem:
                        ret = "r";
                        break;
                    case TYPE.codeseg:
                        ret = "c";
                        break;
                    case TYPE.akey:
                        ret = "ak";
                        break;
                    case TYPE.skey:
                        ret = "sk";
                        break;
                    case TYPE.list:
                        {//Ordered
                            ret = "list(";
                            bool first = true;
                            foreach (GenericData gd in list)
                            {
                                if (!first) ret += ",";
                                ret += gd.TypeString;
                                first = false;
                            }
                            ret += ")";
                            break;
                        }
                    case TYPE.ciphertextbyak:
                        ret = "ciphera(" + plaintext.TypeString + ",ak)";
                        break;
                    case TYPE.ciphertextbysk:
                        ret = "ciphers(" + plaintext.TypeString + ",sk)";
                        break;
                    case TYPE.signature:
                        ret = "sign(" + plaintext.TypeString + ",ak)";
                        break;
                    case TYPE.hash:
                        ret = "hash(" + plaintext.TypeString + ")";
                        break;
                    case TYPE.xor:
                        {//Unordered
                            ret = "xor(";
                            List<string> sublist = new List<string>();
                            foreach (GenericData gd in list)
                                sublist.Add(gd.TypeString);
                            sublist.Sort();
                            bool first = true;
                            foreach (string s in sublist)
                            {
                                if (!first) ret += ",";
                                ret += s;
                                first = false;
                            }
                            ret += ")";
                            break;
                        }
                    case TYPE.boolean:
                        ret = "b";
                        break;
                    case TYPE.nonce:
                        ret = "n";
                        break;
                    case TYPE.otp:
                        ret = "o";
                        break;
                    case TYPE.var:
                        ret = "v";
                        break;
                    /*
                case TYPE.ret:
                    ret = "v";
                    break;*/
                    default:
                        throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("Unknown type");
                }
                return ret;
            }
        }

        public TYPE getType()
        {
            return type;
        }
    }
}
