﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;


namespace PAT.Lib
{
    class Knowledge : ExpressionValue
    {
        public List<GDBase> list = null;

        public Knowledge()
        {
            list = new List<GDBase>();
        }

        public Knowledge(Knowledge k)
        {
            list = new List<GDBase>();
            foreach (GDBase gd in k.list)
            {
                list.Add(gd.clone());
            }
        }

        public Knowledge clear()
        {
            return new Knowledge();
        }

        public override string ToString()
        {
            return "[" + ExpressionID + "]";
        }

        public override ExpressionValue GetClone()
        {
            return new Knowledge(this);
        }

        public override string ExpressionID
        {
            get
            {
                //list.Sort();
                List<string> stringlist = new List<string>();
                foreach (GDBase gd in list)
                {
                    stringlist.Add(gd.ExpressionID);
                }
                stringlist.Sort();
                string ret = "";
                bool first = true;
                foreach (string s in stringlist)
                {
                    if (!first) ret += ", ";
                    ret += s;
                    first = false;
                }
                return ret;
            }
        }
        /*
        public void debug()
        {
            throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException(ExpressionID);
        }*/
        public bool knows(GDBase gd)
        {
            return list.Contains(gd);
        }

        public bool addKnowledge(GDBase gd)
        {
            if (!knows(gd))
            {
                list.Add(gd.clone());
                return true;
            }
            else
                return false;
        }

        public void deduceKnowledge(Cryptography crypt)
        {
            int c = 0;
            bool changed = false;
            do
            {
                changed = _applyRules(crypt);
                c++;
                if (c > 1000) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("stun in rules");
            } while (changed);
            c = 0;
            do
            {
                changed = _reduceKnowledge();
                c++;
                if (c > 1000) throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException("stun in reduce");
            } while (changed);
        }


        /*
        public bool validByName(int name)
        {
            int num = 0;
            foreach (GenericData gd in list)
            {
                if (gd.name == name) num++;
            }
            return num == 1;
        }

        public GenericData findByName(int name)
        {
            GenericData ret = null;
            if (validByName(name))
            {
                foreach (GenericData gd in list)
                {
                    if (gd.name == name) break;
                }
            }
            return ret;
        }
        */
        //remove duplicated knowledge
        public bool _reduceKnowledge()
        {
            //Console.WriteLine("111");
            bool changed = false;
            //debug();
            List<GDBase> subk = new List<GDBase>();
            foreach (GDBase gdb in list)
            {
                subk.Add(gdb);
            }

            for (int i = list.Count - 1; i >= 0; i--)
            {
                GDBase gdb = list[i];
                subk.Remove(gdb);

                bool dup = gdb.duplicate(subk);
                //TODO: remove xor with 0 element
                if (dup)
                {
                    list.RemoveAt(i);
                    changed = true;
                }
                else
                {
                    subk.Add(gdb);
                }

            }
            return changed;
        }

        //reasoning
        public bool _applyRules(Cryptography crypt)
        {
            bool changed = false;
            //Rule 1.a: asymmetric encryption
            {
                List<GDBase> ciphertextlist = list.FindAll(
                                delegate(GDBase gdb)
                                {
                                    return gdb is CiphertextA;
                                });
                List<GDBase> signaturelist = list.FindAll(
                                delegate(GDBase gdb)
                                {
                                    return gdb is Signature;
                                });
                List<GDBase> akeylist = list.FindAll(
                                delegate(GDBase gdb)
                                {
                                    return gdb is AKey;
                                });

                foreach (AKey key in akeylist)
                {
                    foreach (CiphertextA ciphertext in ciphertextlist)
                    {
                        if (crypt.isDecryptable(ciphertext, key))
                        {

                            //list.Remove(ciphertext);//correct?
                            changed |= addKnowledge(ciphertext.plaintext);
                        }
                    }
                    foreach (Signature signature in signaturelist)
                    {
                        if (crypt.verifySignature(signature, key))
                        {

                            //list.Remove(signature);//correct?
                            changed |= addKnowledge(signature.plaintext);
                        }
                    }
                }
            }
            //Rule 1.b: symmetric encryption
            {
                List<GDBase> ciphertextlist = list.FindAll(
                                delegate(GDBase gdb)
                                {
                                    return gdb is CiphertextS;
                                });
                List<GDBase> keylist = list.FindAll(
                                delegate(GDBase gdb)
                                {
                                    return gdb is SKey;
                                });
                //throw new PAT.Common.Classes.Expressions.ExpressionClass.RuntimeException(ciphertextlist.Count.ToString()+" "+keylist.Count.ToString());
                foreach (CiphertextS ciphertext in ciphertextlist)
                {
                    foreach (SKey key in keylist)
                    {
                        if (crypt.isDecryptable(ciphertext, key))
                        {
                            changed |= addKnowledge(ciphertext.plaintext);
                        }
                    }
                }
            }
            //Rule 2:
            {
                List<GDBase> listofconcat = list.FindAll(
                delegate(GDBase gdb)
                {
                    return gdb is Concat;
                });
                foreach (Concat gdb in listofconcat)
                {
                    foreach (GDBase element in gdb.list)
                    {
                        changed |= addKnowledge(element);
                    }
                }
            }
            
            //Rule 3:
            {

                List<GDBase> listofxor = list.FindAll(
                delegate(GDBase gdb)
                {
                    return gdb is XOR;
                });
                List<GDBase> listclone = list.FindAll(
                delegate(GDBase gdb)
                {
                    return true;
                });
                foreach (XOR xor in listofxor)
                {
                    foreach (GDBase gdb in listclone)
                    {
                        XOR newxor = xor.xor(gdb);
                        if (newxor.size>0 && newxor.size < xor.size)
                        {
                            changed |= addKnowledge(newxor);
                        }

                    }
                }

            }
            return changed;
        }
        /*
        public List<GDBase> generateData(GDBase org)
        {
            List<GDBase> result = new List<GDBase>();
            result = org.generate(list);
            return result;
        }
        */
        public GeneratedMsg generateMsg(GDBase gdb)
        {
            GeneratedMsg gm = new GeneratedMsg();
            gm.list = gdb.generate(list);
            //gm.list.Remove(gd);
            //gm.list.Insert(0, new GenericData(gd));
            return gm;
        }
    }

    class GeneratedMsg : ExpressionValue
    {
        public List<GDBase> list = null;

        public GeneratedMsg()
        {
            list = new List<GDBase>();
        }

        public GeneratedMsg(GeneratedMsg gm)
        {
            list=gm.list;

            //list = new List<GDBase>(gm.list);
            /*
            foreach (GDBase gdb in gm.list)
            {
                list.Add(gdb.clone());
            }
             */
        }

        public GeneratedMsg clear()
        {
            return new GeneratedMsg();
        }

        public GDBase at(int i)
        {
            return list[i];
        }

        public int getCount()
        {
            return list.Count;
        }

        public override string ToString()
        {
            return "[" + ExpressionID + "]";
        }

        public override ExpressionValue GetClone()
        {
            return new GeneratedMsg(this);
        }

        public override string ExpressionID
        {
            get
            {
                //CRACK: suppose only one GM in CSP
                if (list == null)
                    return "";
                else
                    return list.Count.ToString();
                /*
                StringBuilder sb = new StringBuilder();
                bool first = true;
                foreach (GDBase gd in list)
                {
                    //if (!first) ret += "\n";
                    if (!first) sb.Append("\n");
                    //ret += s;
                    sb.Append(gd.ExpressionID);
                    first = false;
                }
                return sb.ToString();
                 * */
            }
        }
    }
}
