﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PAT.Lib
{
    class SyncException : System.Exception
    {

        public enum SyncReason { Nothing, RunCode, Halt, VMEntry, VMExit, LL, ChannelInput, ChannelOutput };
        //public enum ErrorReason { InvalidVA, InvalidPA, InvalidVMID, InvalidPT, InvalidReg, InvalidCode, InvalidDataFormat, HaltVM, InvalidSrcEncode, InvalidDstEncode, NoAccess, HostInstruction, Inconsistency, InsufficientResource };
        //public enum VMExitReason { Crash, HYPERCALL, VON, VOFF, VMLAUNCH, VMRESUME, NEWVM, LL };
        
        //public enum CIReason { Device, MOVEsrc };
        //public enum COReason { Device };

        SyncReason _state = SyncReason.Nothing;
        Concat _param = null;
        //Concat _subparam = null;

        public SyncReason state { get { return _state; } }
        public Concat param { get { return _param; } }
        //public Concat subparam { get { return _subparam; } }

        public SyncException(SyncReason __state, Concat __param=null)
        {
            _state = __state;
            _param = __param;
            //_subparam = __subparam;
        }

        public static SyncException syncNothing()
        {
            return new SyncException(SyncReason.Nothing);
        }

        public static SyncException syncRunCode()
        {
            return new SyncException(SyncReason.RunCode);
        }
        /*
        public static SyncException syncError(int reason)
        {
            return new SyncException(SyncState.Error, new Concat(new Value(reason)));
        }
        */
        public static SyncException syncHalt()
        {
            return new SyncException(SyncReason.Halt, new Concat(new Value(0)));
        }

        public static SyncException syncLL()
        {
            return new SyncException(SyncReason.LL, new Concat(new Value(0)));
        }

        public static SyncException syncChannelInput(int chn)
        {
            return new SyncException(SyncReason.ChannelInput, new Concat(new Value(chn)));
        }

        public static SyncException syncChannelInput(int chn, GDBase dst)
        {
            return new SyncException(SyncReason.ChannelInput, new Concat(new Value(chn), dst.clone()));
        }

        public static SyncException syncChannelOutput(int chn, GDBase data)
        {
            return new SyncException(SyncReason.ChannelOutput, new Concat(new Value(chn), data.clone()));
        }

        public static SyncException syncVMEntry()
        {
            return new SyncException(SyncReason.VMEntry);
        }
      
        public static SyncException syncVMExit(Concat param)
        {
            return new SyncException(SyncReason.VMExit, (Concat)param.clone());
        }

        public static SyncException syncVMExit_Error()
        {
            return syncVMExit(new Concat(new Value((int)Processor.VMExitReason.Error)));
        }

        public static SyncException syncVMExit_PriviledgedInstruction()
        {
            return syncVMExit(new Concat(new Value((int)Processor.VMExitReason.PrivilegedInstruction)));
        }
        /*
        public static SyncException syncVMExit_Crash(Concat crashparam)
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.Crash)), crashparam);
        }

        public static SyncException syncVMExit_HYPERCALL(int callid, Concat userparam)
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.HYPERCALL), new Value(callid)), userparam);
        }

        public static SyncException syncVMExit_VON()
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.VON)));
        }

        public static SyncException syncVMExit_VOFF()
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.VOFF)));
        }

        public static SyncException syncVMExit_VMRESUME()
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.VMRESUME)));
        }

        public static SyncException syncVMExit_NEWVM()
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.NEWVM)));
        }

        public static SyncException syncVMExit_LL()
        {
            return new SyncException(SyncState.VMExit, new Concat(new Value((int)VMExitReason.LL)));
        }

        public static SyncException syncChannelInput(int chn)
        {
            return new SyncException(SyncState.ChannelInput, new Concat(new Value(chn)));
        }

        public static SyncException syncChannelOutput(int chn)
        {
            return new SyncException(SyncState.ChannelOutput, new Concat(new Value(chn)));
        }

        public static SyncException syncCrash_InvalidVA(int va)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidVA), new Value(va)));
        }

        public static SyncException syncCrash_InvalidPA(int pa)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidPA), new Value(pa)));
        }

        public static SyncException syncCrash_InvalidVMID(int vmid)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidVMID), new Value(vmid)));
        }

        public static SyncException syncCrash_InvalidPT()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidPT)));
        }

        public static SyncException syncCrash_InvalidDataFormat()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidDataFormat)));
        }

        public static SyncException syncCrash_InvalidReg(Reg reg)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidReg), new Reg(reg)));
        }

        public static SyncException syncCrash_InvalidCode()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidCode)));
        }

        public static SyncException syncCrash_HaltVM(int vmid)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.HaltVM), new Value(vmid)));
        }

        public static SyncException syncCrash_InvalidSrcEncode(GDBase obj)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidSrcEncode), obj.clone()));
        }

        public static SyncException syncCrash_InvalidDstEncode(GDBase obj)
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InvalidDstEncode), obj.clone()));
        }

        public static SyncException syncCrash_NoAccess()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.NoAccess)));
        }

        public static SyncException syncCrash_HostInstruction()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.HostInstruction)));
        }

        public static SyncException syncCrash_Inconsistency()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.Inconsistency)));
        }

        public static SyncException syncCrash_InsufficientResource()
        {
            return new SyncException(SyncState.Crash, new Concat(new Value((int)CrashReason.InsufficientResource)));
        }
        */

    }
}
